Imports System.Windows.Forms

Public Class frmBookmarkEditor

    Private m_bookmark As BookmarkEntry
    Private m_bookman As BookmarkManager

    Public Sub Init(ByVal bm As BookmarkManager, ByVal bookmark As BookmarkEntry, ByVal folder As FolderEntry)
        bm.RootFolder.displayHierarchy(Me.tvHierarchy, folder)

        txtAddress.Text = bookmark.URL
        txtDescription.Text = bookmark.Description
        txtName.Text = bookmark.Name
        chkStarred.Checked = bookmark.Starred

        m_bookmark = bookmark
        m_bookman = bm
    End Sub

    Private Sub cmdOK_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdOK.Click
        Try
            If Not txtAddress.Text.Contains("://") Then
                txtAddress.Text = "http://" & txtAddress.Text
            End If
            Dim url As New Uri(txtAddress.Text)
        Catch ex As UriFormatException
            txtAddress.Focus()
            ttInvalid.Show("You must enter a valid url location", txtAddress, 5000)
            Exit Sub
        End Try

        Try
            If txtName.Text = "" Then
                Throw New IO.InvalidDataException("You must enter a valid name")
            End If

            If txtName.Text.ToLower <> m_bookmark.Name.ToLower Then
                Dim bookmark As BookmarkEntry
                For Each bookmark In CType(tvHierarchy.SelectedNode.Tag, FolderEntry).BookmarkEntries
                    If bookmark.Name.ToLower = txtName.Text.ToLower Then
                        Throw New DuplicateNameException("A bookmark by this name is already in the selected folder." & vbCrLf & _
                              "Select a different folder or change the name.")
                    End If
                Next
            End If
        Catch ex As Exception
            txtName.Focus()
            ttInvalid.Show(ex.Message, txtName, 5000)
            Exit Sub
        End Try

        m_bookmark.Name = txtName.Text
        m_bookmark.Description = txtDescription.Text
        m_bookmark.URL = txtAddress.Text
        m_bookmark.Starred = chkStarred.Checked

        If Not Object.ReferenceEquals(CType(tvHierarchy.SelectedNode.Tag, FolderEntry), m_bookmark.Parent) Then
            If Not m_bookmark.Parent Is Nothing Then
                m_bookmark.Parent.deleteBookmarkEntry(m_bookmark)
            End If
            CType(tvHierarchy.SelectedNode.Tag, FolderEntry).addBookmarkEntry(m_bookmark)
        End If

        Me.DialogResult = System.Windows.Forms.DialogResult.OK
        Me.Close()
    End Sub

    Private Sub cmdCancel_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdCancel.Click
        Me.DialogResult = System.Windows.Forms.DialogResult.Cancel
        Me.Close()
    End Sub

    Private Sub BookmarkEditor_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Me.tvHierarchy.ImageList = frmMain.tvHierarchy.ImageList
    End Sub

    Private Sub cmdSelectFolder_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdSelectFolder.Click
        Me.Width = 455
        Me.tvHierarchy.Show()
    End Sub

    Private Sub tvHierarchy_AfterSelect(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles tvHierarchy.AfterSelect
        Me.txtFolder.Text = CType(e.Node.Tag, FolderEntry).Path
        Me.txtFolder.SelectionStart = Len(Me.txtFolder.Text)
    End Sub

    Private Sub cmdNewFolder_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdNewFolder.Click
        Dim newFolderEditor As New frmFolderEditor
        Dim newFolder As New FolderEntry

        newFolderEditor.Init(m_bookman, newFolder, CType(tvHierarchy.SelectedNode.Tag, FolderEntry))

        If newFolderEditor.ShowDialog() = Windows.Forms.DialogResult.OK Then
            ' add the new folder to this tree
            Dim parentNode As TreeNode
            parentNode = frmMain.FindNode(tvHierarchy, newFolder.Parent)

            Dim newNode As TreeNode
            newNode = New TreeNode(newFolder.Name)
            newNode.Tag = newFolder

            parentNode.Nodes.Add(newNode)
            tvHierarchy.SelectedNode = newNode

            ' now add the new folder to the main list
            parentNode = frmMain.FindNode(frmMain.tvHierarchy, newFolder.Parent)
            newNode = newNode.Clone  ' create a clone because the same node can't be on two trees at the same time

            parentNode.Nodes.Add(newNode)
        End If
    End Sub
End Class
