Public Class frmMain
    ' create a new BookmarkManager
    Friend m_BookmarkManager As New BookmarkManager
    Private m_currFolder As FolderEntry

    Private Sub frmMain_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        m_BookmarkManager.saveEntries("bookmarks.xml")
    End Sub

    Private Sub frmMain_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        tvHierarchy.Sorted = True

        m_BookmarkManager.getEntries("bookmarks.xml")         ' load the bookmarks file

        ' Display the hierarchy for the root folder on tvHierarchy
        m_BookmarkManager.RootFolder.displayHierarchy(Me.tvHierarchy, Nothing)

    End Sub

#Region " tvHierarchy "
    Private Sub tvHierarchy_AfterSelect(ByVal sender As System.Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles tvHierarchy.AfterSelect
        m_currFolder = CType(e.Node.Tag, FolderEntry)

        With m_currFolder
            Dim folder As FolderEntry
            Dim bookmark As BookmarkEntry

            lvBookmarks.Items.Clear()
            For Each folder In .SubFolders
                AddFolderEntry(folder)
            Next
            For Each bookmark In .BookmarkEntries
                AddBookmarkEntry(bookmark)
            Next
        End With
    End Sub

    Private Sub tvHierarchy_DragDrop(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles tvHierarchy.DragDrop
        Dim node As TreeNode, newParent As FolderEntry
        Dim pt As Point = tvHierarchy.PointToClient(New Point(e.X, e.Y))

        node = tvHierarchy.GetNodeAt(pt.X, pt.Y)
        newParent = CType(node.Tag, FolderEntry)

        If e.Data.GetDataPresent(GetType(ListView.SelectedListViewItemCollection)) Then
            Dim draggedItems As ListView.SelectedListViewItemCollection
            draggedItems = e.Data.GetData(GetType(ListView.SelectedListViewItemCollection))

            Dim dragItem As ListViewItem

            For Each dragItem In draggedItems
                If TypeName(dragItem.Tag) = "BookmarkEntry" Then
                    Dim dragBookmark As BookmarkEntry
                    dragBookmark = CType(dragItem.Tag, BookmarkEntry)

                    Dim bookmark As BookmarkEntry
                    For Each bookmark In newParent.BookmarkEntries
                        If bookmark.Name = dragBookmark.Name Then
                            MessageBox.Show("A bookmark named """ & bookmark.Name & """ already exists in the destination folder")
                            Exit Sub
                        End If
                    Next

                    If Not Object.ReferenceEquals(dragBookmark.Parent, newParent) Then
                        dragItem.Remove()

                        m_BookmarkManager.moveBookmark(dragBookmark, newParent)
                    End If
                ElseIf TypeName(dragItem.Tag) = "FolderEntry" Then
                    Dim dragFolder As FolderEntry
                    dragFolder = CType(dragItem.Tag, FolderEntry)

                    Dim folder As FolderEntry
                    For Each folder In newParent.SubFolders
                        If folder.Name = dragFolder.Name Then
                            MessageBox.Show("A folder named """ & folder.Name & """ already exists in the destination folder.")
                            Exit Sub
                        End If
                    Next

                    If Not Object.ReferenceEquals(dragFolder.Parent, newParent) Then
                        If Not Object.ReferenceEquals(dragFolder, newParent) Then
                            If Not newParent.Path.StartsWith(dragFolder.Path) Then
                                FindNode(tvHierarchy, dragFolder).Remove()
                                dragItem.Remove()

                                m_BookmarkManager.moveFolder(dragFolder, newParent)

                                FolderEntry.displayHierarchy(node.Nodes, dragFolder, Nothing)
                            End If
                        End If
                    End If
                End If
            Next
        ElseIf e.Data.GetDataPresent(GetType(TreeNode)) Then
            Dim dragNode As TreeNode
            Dim dragFolder As FolderEntry

            dragNode = e.Data.GetData(GetType(TreeNode))
            dragFolder = CType(dragNode.Tag, FolderEntry)

            Dim folder As FolderEntry
            For Each folder In newParent.SubFolders
                If folder.Name = dragFolder.Name Then
                    MessageBox.Show("A folder named """ & folder.Name & """ already exists in the destination folder.")
                    Exit Sub
                End If
            Next

            If Not Object.ReferenceEquals(dragFolder.Parent, newParent) Then
                If Not Object.ReferenceEquals(dragFolder, newParent) Then
                    If Not newParent.Path.StartsWith(dragFolder.Path) Then
                        dragNode.Remove()

                        m_BookmarkManager.moveFolder(dragFolder, newParent)

                        node.Nodes.Add(dragNode)
                        node.Expand()

                        tvHierarchy.SelectedNode = dragNode
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub tvHierarchy_DragOver(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles tvHierarchy.DragOver
        Dim node As TreeNode
        Dim pt As Point = tvHierarchy.PointToClient(New Point(e.X, e.Y))

        node = tvHierarchy.GetNodeAt(pt.X, pt.Y)
        node.Expand()

        If Not node Is Nothing Then
            If e.Data.GetDataPresent(GetType(TreeNode)) Then
                Dim dragNode As TreeNode
                dragNode = e.Data.GetData(GetType(TreeNode))

                If Object.ReferenceEquals(dragNode.Tag, node.Tag) Then
                    e.Effect = DragDropEffects.None
                    Exit Sub
                End If

                If CType(node.Tag, FolderEntry).Path.StartsWith(CType(dragNode.Tag, FolderEntry).Path) Then
                    e.Effect = DragDropEffects.None
                    Exit Sub
                End If

                If Object.ReferenceEquals(CType(dragNode.Tag, FolderEntry).Parent, node.Tag) Then
                    e.Effect = DragDropEffects.None
                    Exit Sub
                End If
            End If
            If e.Data.GetDataPresent(GetType(ListView.SelectedListViewItemCollection)) Then
                Dim draggedItems As ListView.SelectedListViewItemCollection
                Dim dragItem As ListViewItem

                draggedItems = e.Data.GetData(GetType(ListView.SelectedListViewItemCollection))

                For Each dragItem In draggedItems
                    If TypeName(dragItem.Tag) = "FolderEntry" Then
                        If Object.ReferenceEquals(dragItem.Tag, node.Tag) Then
                            e.Effect = DragDropEffects.None
                            Exit Sub
                        End If

                        If CType(node.Tag, FolderEntry).Path.StartsWith(CType(dragItem.Tag, FolderEntry).Path) Then
                            e.Effect = DragDropEffects.None
                            Exit Sub
                        End If

                        If Object.ReferenceEquals(CType(dragItem.Tag, FolderEntry).Parent, node.Tag) Then
                            e.Effect = DragDropEffects.None
                            Exit Sub
                        End If
                    Else
                        If Object.ReferenceEquals(CType(dragItem.Tag, BookmarkEntry).Parent, node.Tag) Then
                            e.Effect = DragDropEffects.None
                            Exit Sub
                        End If
                    End If
                Next
            End If

            e.Effect = DragDropEffects.Move
        Else
            e.Effect = DragDropEffects.None
        End If
    End Sub

    Private Sub tvHierarchy_ItemDrag(ByVal sender As Object, ByVal e As System.Windows.Forms.ItemDragEventArgs) Handles tvHierarchy.ItemDrag
        pnlRoot.Visible = True
        DoDragDrop(e.Item, DragDropEffects.Move)
        pnlRoot.Visible = False
    End Sub

    Private Sub tvHierarchy_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles tvHierarchy.KeyDown
        If e.KeyCode = Keys.Delete Then
            mnuHierDelete.PerformClick()
        End If
    End Sub

    Private Sub tvHierarchy_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles tvHierarchy.MouseDown
        Dim node As TreeNode
        node = tvHierarchy.GetNodeAt(e.X, e.Y)

        tvHierarchy.SelectedNode = node

        If node Is Nothing Then
            m_currFolder = m_BookmarkManager.RootFolder
            ssStatus.Items("Location").Text = "Contains " & m_BookmarkManager.RootFolder.SubFolders.Count.ToString & " Folders"
        Else
            m_currFolder = CType(node.Tag, FolderEntry)
            Dim folder As FolderEntry
            folder = CType(node.Tag, FolderEntry)
            ssStatus.Items("Location").Text = "Contains " & m_currFolder.SubFolders.Count.ToString & " Folders and " & m_currFolder.BookmarkEntries.Count.ToString & " Bookmarks"
        End If
    End Sub
#End Region

#Region " lvBookmarks "
    Private Sub lvBookmarks_AfterLabelEdit(ByVal sender As Object, ByVal e As System.Windows.Forms.LabelEditEventArgs) Handles lvBookmarks.AfterLabelEdit
        If e.Label Is Nothing Then Return
        Select Case TypeName(lvBookmarks.Items(e.Item).Tag)
            Case "FolderEntry"
                Dim folder As FolderEntry
                folder = CType(lvBookmarks.Items(e.Item).Tag, FolderEntry)
                FindNode(tvHierarchy, folder).Text = e.Label

                folder.Name = e.Label
            Case "BookmarkEntry"
                CType(lvBookmarks.Items(e.Item).Tag, BookmarkEntry).Name = e.Label
        End Select
    End Sub

    Private Sub lvBookmarks_DoubleClick(ByVal sender As Object, ByVal e As System.EventArgs) Handles lvBookmarks.DoubleClick
        With lvBookmarks.FocusedItem
            If .Group.Name = "Folder" Then
                Dim node As TreeNode
                For Each node In tvHierarchy.SelectedNode.Nodes
                    If Object.ReferenceEquals(.Tag, node.Tag) Then
                        tvHierarchy.SelectedNode = node
                        Exit For
                    End If
                Next
            Else
                Process.Start(CType(.Tag, BookmarkEntry).URL)
            End If
        End With
    End Sub

    Private Sub lvBookmarks_DragDrop(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles lvBookmarks.DragDrop
        Dim item As ListViewItem, newParent As FolderEntry
        Dim pt As Point = lvBookmarks.PointToClient(New Point(e.X, e.Y))

        item = lvBookmarks.GetItemAt(pt.X, pt.Y)
        newParent = CType(item.Tag, FolderEntry)

        If e.Data.GetDataPresent(GetType(ListView.SelectedListViewItemCollection)) Then
            Dim draggedItems As ListView.SelectedListViewItemCollection
            draggedItems = e.Data.GetData(GetType(ListView.SelectedListViewItemCollection))

            Dim dragItem As ListViewItem

            For Each dragItem In draggedItems
                If TypeName(dragItem.Tag) = "BookmarkEntry" Then
                    Dim dragBookmark As BookmarkEntry
                    dragBookmark = CType(dragItem.Tag, BookmarkEntry)

                    Dim bookmark As BookmarkEntry
                    For Each bookmark In newParent.BookmarkEntries
                        If bookmark.Name = dragBookmark.Name Then
                            MessageBox.Show("A bookmark named """ & bookmark.Name & """ already exists in the destination folder")
                            Exit Sub
                        End If
                    Next

                    If Not Object.ReferenceEquals(dragBookmark.Parent, newParent) Then
                        dragItem.Remove()

                        m_BookmarkManager.moveBookmark(dragBookmark, newParent)
                    End If
                ElseIf TypeName(dragItem.Tag) = "FolderEntry" Then
                    Dim dragFolder As FolderEntry
                    dragFolder = CType(dragItem.Tag, FolderEntry)

                    Dim folder As FolderEntry
                    For Each folder In newParent.SubFolders
                        If folder.Name = dragFolder.Name Then
                            MessageBox.Show("A folder named """ & folder.Name & """ already exists in the destination folder.")
                            Exit Sub
                        End If
                    Next

                    If Not Object.ReferenceEquals(dragFolder.Parent, newParent) Then
                        If Not Object.ReferenceEquals(dragFolder, newParent) Then
                            Dim oldNode As TreeNode
                            oldNode = FindNode(tvHierarchy, dragFolder)
                            oldNode.Remove()

                            dragItem.Remove()

                            m_BookmarkManager.moveFolder(dragFolder, newParent)

                            FindNode(tvHierarchy, newParent).Nodes.Add(oldNode)
                        End If
                    End If
                End If
            Next
        End If
    End Sub

    Private Sub lvBookmarks_DragOver(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles lvBookmarks.DragOver
        Dim item As ListViewItem
        Dim pt As Point = lvBookmarks.PointToClient(New Point(e.X, e.Y))
        item = lvBookmarks.GetItemAt(pt.X, pt.Y)

        e.Effect = DragDropEffects.None

        If e.Data.GetDataPresent(GetType(ListView.SelectedListViewItemCollection)) Then
            Dim draggedItems As ListView.SelectedListViewItemCollection
            Dim dragItem As ListViewItem

            draggedItems = e.Data.GetData(GetType(ListView.SelectedListViewItemCollection))

            For Each dragItem In draggedItems
                If TypeName(item.Tag) = "FolderEntry" Then
                    If Object.ReferenceEquals(dragItem, item) Then
                        Exit Sub
                    End If
                Else
                    Exit Sub ' exit so effect is None
                End If
            Next
        End If

        e.Effect = DragDropEffects.Move
    End Sub

    Private Sub lvBookmarks_ItemDrag(ByVal sender As Object, ByVal e As System.Windows.Forms.ItemDragEventArgs) Handles lvBookmarks.ItemDrag
        Dim DataObj As New DataObject, itemTypeFlag As Integer
        DataObj.SetData(lvBookmarks.SelectedItems)

        Dim item As ListViewItem
        For Each item In lvBookmarks.SelectedItems
            If item.Group.Name = "Folder" Then
                itemTypeFlag = itemTypeFlag Or 1
            Else
                itemTypeFlag = itemTypeFlag Or 2
            End If
            If itemTypeFlag = 3 Then
                Exit For
            End If
        Next
        DataObj.SetData("ItemTypes", itemTypeFlag)
        If itemTypeFlag = 1 Then
            pnlRoot.Visible = True
        End If

        DoDragDrop(DataObj, DragDropEffects.Move)

        pnlRoot.Visible = False
    End Sub

    Private Sub lvBookmarks_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles lvBookmarks.KeyDown
        If e.KeyCode = Keys.Delete Then
            cmnuDelete.PerformClick()
        ElseIf e.KeyCode = Keys.Enter Then
            lvBookmarks_DoubleClick(sender, e)
        End If
    End Sub

    Private Sub lvBookmarks_SelectedIndexChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles lvBookmarks.SelectedIndexChanged
        If lvBookmarks.SelectedItems.Count = 1 Then
            If TypeName(lvBookmarks.SelectedItems(0).Tag) = "BookmarkEntry" Then
                ssStatus.Items("Location").Text = CType(lvBookmarks.SelectedItems(0).Tag, BookmarkEntry).URL
            Else
                Dim folder As FolderEntry = CType(lvBookmarks.SelectedItems(0).Tag, FolderEntry)
                ssStatus.Items("Location").Text = "Contains " & folder.SubFolders.Count.ToString & " Folders and " & folder.BookmarkEntries.Count.ToString & " Bookmarks"
            End If
        ElseIf lvBookmarks.SelectedItems.Count > 1 Then
            ssStatus.Items("Location").Text = lvBookmarks.SelectedItems.Count.ToString & " Items selected"
        Else
            ssStatus.Items("Location").Text = ""
        End If
    End Sub

#End Region

    Private Sub AddFolderEntry(ByVal folder As FolderEntry)
        Dim newItem As ListViewItem
        newItem = New ListViewItem(folder.Name)
        newItem.Tag = folder
        newItem.Group = lvBookmarks.Groups("Folder")
        newItem.ImageKey = "Folder"

        lvBookmarks.Items.Add(newItem)
    End Sub

    Private Sub AddBookmarkEntry(ByVal bookmark As BookmarkEntry)
        Dim newItem As ListViewItem
        newItem = New ListViewItem(bookmark.Name)
        newItem.SubItems.Add(bookmark.URL)
        newItem.SubItems.Add(bookmark.Description)

        newItem.Tag = bookmark
        If bookmark.Starred Then
            newItem.Group = lvBookmarks.Groups("FavBook")
        Else
            newItem.Group = lvBookmarks.Groups("Bookmark")
        End If
        newItem.ImageKey = "Bookmark"

        lvBookmarks.Items.Add(newItem)
    End Sub

    'finds the node that represents the FolderEntry
    Public Shared Function FindNode(ByVal tvHierarchy As TreeView, ByVal folder As FolderEntry) As TreeNode
        Dim pathParts As String(), pathPart As String
        Dim nodes As TreeNodeCollection, node As TreeNode
        pathParts = folder.Path.Trim("/"c).Split("/"c)
        nodes = tvHierarchy.Nodes

        FindNode = Nothing
        For Each pathPart In pathParts
            For Each node In nodes
                If node.Text = pathPart Then
                    FindNode = node
                    nodes = node.Nodes
                    Exit For
                End If
            Next
        Next
        Return FindNode
    End Function

#Region " Menu Click Handlers "
    Private Sub NewBookmarkItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tsbAddNewBookmark.Click, mnuFileNewBookmark.Click, mnuHierNewBookmark.Click
        Dim newBookmarkEditor As New frmBookmarkEditor
        Dim newBookmark As New BookmarkEntry

        newBookmarkEditor.Init(m_BookmarkManager, newBookmark, m_currFolder)

        If newBookmarkEditor.ShowDialog() = Windows.Forms.DialogResult.OK Then
            If Object.ReferenceEquals(newBookmark.Parent, m_currFolder) Then
                AddBookmarkEntry(newBookmark)
            End If
        End If
    End Sub

    Private Sub NewFolderItem_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tsbAddNewFolder.Click, mnuFileNewFolder.Click, mnuHierNewFolder.Click
        Dim newFolderEditor As New frmFolderEditor
        Dim newFolder As New FolderEntry

        newFolderEditor.Init(m_BookmarkManager, newFolder, m_currFolder)

        If newFolderEditor.ShowDialog() = Windows.Forms.DialogResult.OK Then
            If Object.ReferenceEquals(newFolder.Parent, m_currFolder) Then
                Dim newNode As TreeNode
                newNode = New TreeNode(newFolder.Name)
                newNode.Tag = newFolder
                tvHierarchy.SelectedNode.Nodes.Add(newNode)

                AddFolderEntry(newFolder)
            Else
                Dim newNode As TreeNode, parentNode As TreeNode
                parentNode = FindNode(tvHierarchy, newFolder.Parent)
                newNode = New TreeNode(newFolder.Name)
                newNode.Tag = newFolder
                If Not parentNode Is Nothing Then
                    parentNode.Nodes.Add(newNode)
                Else
                    tvHierarchy.Nodes.Add(newNode)
                End If
                tvHierarchy.SelectedNode = newNode
            End If
        End If
    End Sub

    Private Sub mnuHelpAbout_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuHelpAbout.Click
        frmAbout.ShowDialog()
    End Sub

    Private Sub mnuFileSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileSave.Click
        m_BookmarkManager.saveEntries("bookmarks.xml")
    End Sub

    Private Sub mnuFileSaveAs_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileSaveAs.Click
        Dim sfdSave As New SaveFileDialog
        sfdSave.Filter = "XML Files|*.xml"
        sfdSave.AddExtension = True
        sfdSave.DefaultExt = "xml"

        If sfdSave.ShowDialog() = Windows.Forms.DialogResult.OK Then
            m_BookmarkManager.saveEntries(sfdSave.FileName)
        End If
    End Sub

    Private Sub mnuFileOpen_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileOpen.Click
        Dim ofdOpen As New OpenFileDialog
        ofdOpen.Filter = "XML Files|*.xml"
        ofdOpen.AddExtension = True
        ofdOpen.DefaultExt = "xml"
        If ofdOpen.ShowDialog = Windows.Forms.DialogResult.OK Then
            m_BookmarkManager = New BookmarkManager
            m_BookmarkManager.getEntries(ofdOpen.FileName)

            tvHierarchy.Nodes.Clear()

            m_BookmarkManager.RootFolder.displayHierarchy(tvHierarchy, Nothing)
        End If
    End Sub

    Private Sub mnuView_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewStatusbar.Click
        mnuViewStatusbar.Checked = Not mnuViewStatusbar.Checked
        ssStatus.Visible = mnuViewStatusbar.Checked
    End Sub

    Private Sub mnuViewRefresh_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewRefresh.Click
        Dim tn As TreeNode
        tn = tvHierarchy.SelectedNode
        tvHierarchy.SelectedNode = Nothing
        tvHierarchy.SelectedNode = tn
        tn = Nothing
    End Sub

    Private Sub mnuFileExit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileExit.Click
        Me.Close()
    End Sub

    Private Sub tsbAddNew_DropDownOpening(ByVal sender As Object, ByVal e As System.EventArgs) Handles tsbAddNew.DropDownOpening
        tsbAddNewBookmark.Enabled = Not (tvHierarchy.SelectedNode Is Nothing)
    End Sub

    Private Sub mnuFileNew_DropDownOpening(ByVal sender As Object, ByVal e As System.EventArgs) Handles mnuFileNew.DropDownOpening
        mnuFileNewBookmark.Enabled = Not (tvHierarchy.SelectedNode Is Nothing)
    End Sub

#Region " Context Menus "
    Private Sub cmnuBookmarks_Opening(ByVal sender As Object, ByVal e As System.ComponentModel.CancelEventArgs) Handles cmnuBookmarks.Opening
        cmnuEdit.Enabled = lvBookmarks.SelectedItems.Count = 1
    End Sub

    Private Sub cmnuEdit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tbsEdit.Click, cmnuEdit.Click
        If lvBookmarks.SelectedItems.Count = 1 Then
            If TypeName(lvBookmarks.SelectedItems(0).Tag) = "FolderEntry" Then
                Dim newFolderEditor As New frmFolderEditor
                Dim folder As FolderEntry
                Dim oldParent As FolderEntry
                Dim node As TreeNode

                folder = CType(lvBookmarks.SelectedItems(0).Tag, FolderEntry)
                oldParent = folder.Parent
                newFolderEditor.Init(m_BookmarkManager, folder, oldParent)

                node = FindNode(tvHierarchy, folder)

                If newFolderEditor.ShowDialog() = Windows.Forms.DialogResult.OK Then
                    If Object.ReferenceEquals(folder.Parent, oldParent) Then
                        With lvBookmarks.SelectedItems(0)
                            .Text = folder.Name

                            node.Text = folder.Name
                        End With
                    Else
                        lvBookmarks.SelectedItems(0).Remove()

                        node.Remove()

                        If Object.ReferenceEquals(folder.Parent, m_BookmarkManager.RootFolder) Then
                            tvHierarchy.Nodes.Add(node)
                        Else
                            FindNode(tvHierarchy, folder.Parent).Nodes.Add(node)
                        End If
                    End If
                End If
            Else
                Dim newBookmarkEditor As New frmBookmarkEditor
                Dim bookmark As BookmarkEntry
                Dim oldParent As FolderEntry

                bookmark = CType(lvBookmarks.SelectedItems(0).Tag, BookmarkEntry)
                oldParent = bookmark.Parent
                newBookmarkEditor.Init(m_BookmarkManager, bookmark, oldParent)

                If newBookmarkEditor.ShowDialog() = Windows.Forms.DialogResult.OK Then
                    If Object.ReferenceEquals(bookmark.Parent, oldParent) Then
                        With lvBookmarks.SelectedItems(0)
                            .Text = bookmark.Name
                            .SubItems(1).Text = bookmark.URL
                            .SubItems(2).Text = bookmark.Description
                            If bookmark.Starred Then
                                .Group = lvBookmarks.Groups("FavBook")
                            Else
                                .Group = lvBookmarks.Groups("Bookmark")
                            End If
                        End With
                    Else
                        lvBookmarks.SelectedItems(0).Remove()
                    End If
                End If
            End If
        End If
    End Sub

    Private Sub cmnuDelete_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmnuDelete.Click
        If MessageBox.Show("Are you sure you want to delete the selected items?", "Deleting", _
                    MessageBoxButtons.YesNo, MessageBoxIcon.Question) = Windows.Forms.DialogResult.Yes Then

            Dim item As ListViewItem

            For Each item In lvBookmarks.SelectedItems
                If TypeName(item.Tag) = "BookmarkEntry" Then
                    Dim bookmark As BookmarkEntry
                    bookmark = CType(item.Tag, BookmarkEntry)
                    bookmark.Parent.deleteBookmarkEntry(bookmark)

                    item.Remove()
                Else ' must be a folder
                    Dim folder As FolderEntry
                    folder = CType(item.Tag, FolderEntry)
                    folder.Parent.deleteFolderEntry(folder)

                    item.Remove()

                    For Each node As TreeNode In tvHierarchy.SelectedNode.Nodes
                        If Object.ReferenceEquals(node.Tag, folder) Then
                            node.Remove()
                            Exit For
                        End If
                    Next
                End If
            Next
        End If
    End Sub

    Private Sub cmnuHierarchy_Opening(ByVal sender As Object, ByVal e As System.ComponentModel.CancelEventArgs) Handles cmnuHierarchy.Opening
        mnuHierNewBookmark.Enabled = Not (m_currFolder.Name = "Root")
    End Sub

    Private Sub mnuEditNode_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuHierEdit.Click
        Dim newFolderEditor As New frmFolderEditor
        Dim folder As FolderEntry
        Dim oldParent As FolderEntry
        Dim node As TreeNode

        node = tvHierarchy.SelectedNode
        folder = CType(node.Tag, FolderEntry)
        oldParent = folder.Parent
        newFolderEditor.Init(m_BookmarkManager, folder, folder.Parent)

        If newFolderEditor.ShowDialog = Windows.Forms.DialogResult.OK Then
            node.Text = folder.Name
            If Not Object.ReferenceEquals(folder.Parent, oldParent) Then
                node.Remove()

                If Object.ReferenceEquals(folder.Parent, m_BookmarkManager.RootFolder) Then
                    tvHierarchy.Nodes.Add(node)
                Else
                    FindNode(tvHierarchy, folder.Parent).Nodes.Add(node)
                End If

                tvHierarchy.SelectedNode = node
            End If
        End If
    End Sub

    Private Sub mnuDeleteNode_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuHierDelete.Click
        If MessageBox.Show("Deleting this folder will remove all sub folders and bookmarks within.  Are you sure?", "Deleting", _
                     MessageBoxButtons.YesNo, MessageBoxIcon.Question) = Windows.Forms.DialogResult.Yes Then

            Dim node As TreeNode, folder As FolderEntry
            node = tvHierarchy.SelectedNode
            folder = CType(node.Tag, FolderEntry)
            folder.Parent.deleteFolderEntry(folder)

            node.Remove()
        End If
    End Sub
#End Region
#End Region

#Region " View Controllers "
    Private Sub mnuViewTiles_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewTiles.Click
        RadioCheckView(View.Tile)
    End Sub

    Private Sub mnuViewIcons_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewIcons.Click
        RadioCheckView(View.LargeIcon)
    End Sub

    Private Sub mnuViewList_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewList.Click
        RadioCheckView(View.List)
    End Sub

    Private Sub mnuViewDetails_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuViewDetails.Click
        RadioCheckView(View.Details)
    End Sub

    Private Sub RadioCheckView(ByVal view As Windows.Forms.View)
        mnuViewTiles.Checked = view = Windows.Forms.View.Tile
        mnuViewIcons.Checked = view = Windows.Forms.View.LargeIcon
        mnuViewList.Checked = view = Windows.Forms.View.List
        mnuViewDetails.Checked = view = Windows.Forms.View.Details

        lvBookmarks.View = view
    End Sub
#End Region

    Private Sub pnlRoot_DragDrop(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles pnlRoot.DragDrop
        If e.Data.GetDataPresent(GetType(TreeNode)) Then
            Dim dragNode As TreeNode, dragFolder As FolderEntry
            dragNode = e.Data.GetData(GetType(TreeNode))
            dragFolder = CType(dragNode.Tag, FolderEntry)

            Dim folder As FolderEntry
            For Each folder In m_BookmarkManager.RootFolder.SubFolders
                If folder.Name = dragFolder.Name Then
                    MessageBox.Show("A folder named """ & folder.Name & """ already exists in the destination folder.")
                    Exit Sub
                End If
            Next

            m_BookmarkManager.moveFolder(dragFolder, m_BookmarkManager.RootFolder)
            dragNode.Remove()

            tvHierarchy.Nodes.Add(dragNode)
        ElseIf e.Data.GetDataPresent(GetType(ListView.SelectedListViewItemCollection)) Then
            Dim draggedItems As ListView.SelectedListViewItemCollection
            draggedItems = e.Data.GetData(GetType(ListView.SelectedListViewItemCollection))

            Dim dragItem As ListViewItem

            For Each dragItem In draggedItems
                Dim dragFolder As FolderEntry
                dragFolder = CType(dragItem.Tag, FolderEntry)

                Dim folder As FolderEntry
                For Each folder In m_BookmarkManager.RootFolder.SubFolders
                    If folder.Name = dragFolder.Name Then
                        MessageBox.Show("A folder named """ & folder.Name & """ already exists in the destination folder.")
                        Exit Sub
                    End If
                Next

                Dim node As TreeNode
                node = FindNode(tvHierarchy, dragFolder)
                node.Remove()

                m_BookmarkManager.moveFolder(dragFolder, m_BookmarkManager.RootFolder)
                dragItem.Remove()
                tvHierarchy.Nodes.Add(node)

                tvHierarchy.SelectedNode = node
            Next
        End If
    End Sub

    Private Sub pnlRoot_DragOver(ByVal sender As Object, ByVal e As System.Windows.Forms.DragEventArgs) Handles pnlRoot.DragOver
        e.Effect = DragDropEffects.None
        If e.Data.GetDataPresent(GetType(TreeNode)) Then
            e.Effect = DragDropEffects.Move
        ElseIf e.Data.GetDataPresent(GetType(ListView.SelectedListViewItemCollection)) Then
            e.Effect = DragDropEffects.Move
        End If
    End Sub

    Private Sub Import_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileImportInternetExplorer.Click, mnuFileImportFirefox.Click, mnuFileImportMozilla.Click, mnuFileImportNetscape8.Click
        Dim browserImport As Browser

        Select Case CType(sender, ToolStripMenuItem).Name
            Case mnuFileImportInternetExplorer.Name
                browserImport = New InternetExplorer
            Case mnuFileImportFirefox.Name
                browserImport = New Firefox
            Case mnuFileImportMozilla.Name
                browserImport = New Mozilla
            Case mnuFileImportNetscape8.Name
                browserImport = New Netscape
            Case Else
                Exit Sub
        End Select

        browserImport.import(m_BookmarkManager.RootFolder())
        ' Display the hierarchy for the root folder on tvHierarchy
        m_BookmarkManager.RootFolder.displayHierarchy(Me.tvHierarchy, Nothing)
    End Sub

    Private Sub Export_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles mnuFileExportInternetExplorer.Click, mnuFileExportFirefox.Click, mnuFileExportMozilla.Click, mnuFileExportNetscape8.Click
        Dim browserExport As Browser

        Select Case CType(sender, ToolStripMenuItem).Name
            Case mnuFileExportInternetExplorer.Name
                browserExport = New InternetExplorer
            Case mnuFileExportFirefox.Name
                browserExport = New Firefox
            Case mnuFileExportMozilla.Name
                browserExport = New Mozilla
            Case mnuFileExportNetscape8.Name
                browserExport = New Netscape
            Case Else
                Exit Sub
        End Select

        browserExport.export(m_BookmarkManager.RootFolder())
        ' Display the hierarchy for the root folder on tvHierarchy
        m_BookmarkManager.RootFolder.displayHierarchy(Me.tvHierarchy, Nothing)
    End Sub
End Class
